package org.dyndns.widerstand.proxy.grizzly;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.sun.grizzly.CallbackHandler;
import com.sun.grizzly.Context;
import com.sun.grizzly.Controller;
import com.sun.grizzly.IOEvent;
import com.sun.grizzly.ProtocolFilter;
import com.sun.grizzly.TCPConnectorHandler;
import com.sun.grizzly.TCPSelectorHandler;
import com.sun.grizzly.Controller.Protocol;
import com.sun.grizzly.util.WorkerThread;


public class ProxyFilter implements ProtocolFilter {

    private static final String CHANNEL_MAPPING = "ChannelMapping";

    private final static Logger logger          = Proxy.getLogger();

    private String              remoteServer;

    private int                 remoteServerPort;

    public ProxyFilter(String remoteServer, int remoteServerPort) {
        this.remoteServer = remoteServer;
        this.remoteServerPort = remoteServerPort;
    }

    public boolean execute(Context ctx) throws IOException {
        Controller.logger().log(Level.FINE, "ProxyFilter executing ...");
        ByteBuffer byteBuffer = ((WorkerThread) Thread.currentThread()).getByteBuffer();

        SelectionKey selectionKey1 = ctx.getSelectionKey();
        // get SocketChannel
        SocketChannel channel1 = (SocketChannel) selectionKey1.channel();
        SocketChannel channel2;

        // aquire a ConnectorHandler
        final TCPConnectorHandler tcpConnectorHandler = (TCPConnectorHandler) ctx.getController().acquireConnectorHandler(Controller.Protocol.TCP);

        Map<SocketChannel, SocketChannel> channelMapping = (Map<SocketChannel, SocketChannel>) ctx.getAttribute(CHANNEL_MAPPING);

        if (channelMapping == null) {
            if (logger.isLoggable(Level.FINE)) {
                logger.log(Level.FINE, "creating new channel map ...");
            }
            channelMapping = new HashMap<SocketChannel, SocketChannel>();
            // store Map in Context
            ctx.setAttribute(CHANNEL_MAPPING, channelMapping);
        }

        if (!channelMapping.containsKey(channel1)) {
            if (logger.isLoggable(Level.FINE)) {
                logger.log(Level.FINE, "connecting and creating new channel mapping ...");
            }
            tcpConnectorHandler.connect(new InetSocketAddress(remoteServer, remoteServerPort), new CallbackHandler<Context>() {

                public void onConnect(IOEvent<Context> ioEvent) {
                    if (logger.isLoggable(Level.FINE)) {
                        logger.log(Level.FINE, "got connection ...");
                    }
                    SelectionKey key = ioEvent.attachment().getSelectionKey();
                    tcpConnectorHandler.finishConnect(key);
                    ioEvent.attachment().getController().registerKey(key, SelectionKey.OP_READ, Controller.Protocol.TCP);
                    key.attach(null);
                }

                public void onRead(IOEvent<Context> ioEvent) {

                }

                public void onWrite(IOEvent<Context> ioEvent) {

                }
            }, (TCPSelectorHandler) tcpConnectorHandler.getController().getSelectorHandler(Controller.Protocol.TCP));
            // remember SocketChannel (client proxy) and (proxy server) in Map
            channel2 = tcpConnectorHandler.getUnderlyingChannel();
            channelMapping.put(channel1, channel2);
            // and other direction
            channelMapping.put(tcpConnectorHandler.getUnderlyingChannel(), channel1);
        } else {
            if (logger.isLoggable(Level.FINE)) {
                logger.log(Level.FINE, "using existing channel mapping ...");
            }
            channel2 = channelMapping.get(channel1);
            tcpConnectorHandler.setUnderlyingChannel(channel2);
        }

        if (logger.isLoggable(Level.FINE)) {
            logger.log(Level.FINE, "writing bytes ...");
        }
        // write all out
        tcpConnectorHandler.write(byteBuffer, true);

        // // reregister reading
        // if (logger.isLoggable(Level.FINE)) {
        // logger.log(Level.FINE, "reregister read channel for reading ...");
        // }
        // ctx.getController().registerKey(selectionKey1);
        //        
        // if (logger.isLoggable(Level.FINE)) {
        // logger.log(Level.FINE, "reregister write channel for reading ...");
        // }
        // SelectionKey selectionKey2 =
        // channel2.keyFor(tcpConnectorHandler.getSelectorHandler().getSelector());
        // ctx.getController().registerKey(selectionKey2);

        // release ConnectorHandler;
        ctx.getController().releaseConnectorHandler(tcpConnectorHandler);
        if (logger.isLoggable(Level.FINE)) {
            logger.log(Level.FINE, "ConnectorHandler released.");
        }
        // TODO: callbackhandler is missing ? -> must be null otherwise filters
        // are not executed ...
        return true;
    }

    public boolean postExecute(Context ctx) throws IOException {
        return true;
    }
}
